# TODO [DG] allow user to set default resource types for user, project, or org,
#  including coarse selections like 'local' or 'aws'
# TODO allow registration of user-created resource implementations
